from src.easy_config_hub import ConfigBase, Setting, Level
import pytest


@pytest.fixture
def Config():
    class Config_(ConfigBase):
        version = Setting[str]('0.1.0', 'Version', level=Level.USER | Level.READ_ONLY)
        dev_mode = Setting[bool](True, 'Dev Mode', level=Level.USER)
        debug_mode = Setting[bool](True, 'Debug Mode', level=Level.USER_DEV)
        
        class ImageDownloading(ConfigBase):
            convert_image = Setting[bool](True, 'Convert Image')
            preferable_format = Setting[str]('WEBP', 'Converted Images Format')
    
    return Config_()

def test_config_init():
    class Config_(ConfigBase):
        setting = Setting[str]('difj')
    assert isinstance(Config_(), ConfigBase)
    
def test_nested_config_init(Config):
    assert isinstance(Config, ConfigBase)
    



from __future__ import annotations
import typing
import sys
import os
from pathlib import Path
from .config import ConfigMeta
from .setting import Setting



# I take the main directory, check for child classes, save child classes and initialize them in init
# Do for each dir

def _get_set_std_dir():
    if getattr(sys, 'frozen', False):   # executable mode
        STD_DIR = Path(sys.argv[0]).parent
        print(f'Exe detected, working directory: {STD_DIR}')
    else:   # normal Python environment
        STD_DIR = Path(sys.argv[0]).parent
    STD_DIR = STD_DIR.absolute()
    os.chdir(STD_DIR)
    return STD_DIR

class DirConfigMeta(ConfigMeta):
    STD_DIR: Path = _get_set_std_dir()
    
    def _at_class_creation(cls, target_cls, name):
        target_cls._nested_classes = {}
        target_cls._to_path = {}
    
    def _handle_key_value(cls, target_cls, key, value):
        if isinstance(value, type) and issubclass(value, DirConfigBase):
            target_cls._nested_classes[key] = value
        elif isinstance(value, Path):
            setattr(target_cls, key, value)
        elif isinstance(value, str) and value not in ['__main__', cls.__module__, __file__, target_cls.__name__]:
            target_cls._to_path[key] = value
        elif isinstance(value, ConfigMeta):
            raise Exception('Only Dir configs are allowed')
        

class StdDirConfigBase(metaclass=DirConfigMeta):
    def __init__(self):
        for name, dir_config in self._nested_classes.items():
            instance = dir_config(self.STD_DIR)
            setattr(self.__class__, name, instance)
            
        for name, path_str in self._to_path.items():
            path: Path = self.STD_DIR / path_str
            if path.suffix:
                path.parent.mkdir(exist_ok=True, parents=True)
            else:
                path.mkdir(exist_ok=True, parents=True)
                
            setattr(self.__class__, name, path)
        
    def to_str(self, tabs) -> str:
        s = f"{self._config_name} ({self.STD_DIR}):\n"
        for key, value in self.__class__.__dict__.items():
            if isinstance(value, Path):
                s += f'{'  '*tabs}{key}: {value}\n'
            if isinstance(value, DirConfigBase):
                s += '\n'
                s += '  '*tabs + value.to_str(tabs+1)
        return s
    
class DirConfigBase(metaclass=DirConfigMeta):
    def __init__(self, parent_dir: Path):
        self.parent_dir = parent_dir
        self.dir = parent_dir / self._config_name.lower()
        
        for name, dir_config in self._nested_classes.items():
            instance = dir_config(self.dir)
            setattr(self.__class__, name, instance)
            
        for name, path_str in self._to_path.items():
            path: Path = self.dir / path_str
            if path.suffix:
                path.parent.mkdir(exist_ok=True, parents=True)
            else:
                path.mkdir(exist_ok=True, parents=True)
                
            setattr(self.__class__, name, path)
            
    def to_str(self, tabs) -> str:
        s = f"{self._config_name} ({self.dir}):\n"
        for key, value in self.__class__.__dict__.items():
            if isinstance(value, Path):
                s += f'{'  '*tabs}{key}: {value}\n'
            if isinstance(value, DirConfigBase):
                s += '  '*tabs + value.to_str(tabs+1)
        return s


import multiprocessing
import json
from easy_config_hub.config import ConfigBase
from easy_config_hub.directories_config import StdDirConfigBase, DirConfigBase
from easy_config_hub.enums_config import EnumsConfig, En, a
from easy_config_hub.setting import Setting, VersionSetting, RangeSetting, Level, SettingType


class Config_(ConfigBase):
    version = VersionSetting(0, 1, 0, 'alpha', name='Version', level=Level.USER | Level.READ_ONLY)
    dev_mode = Setting[bool](True, 'Dev Mode', level=Level.USER)
    debug_mode = Setting[bool](True, 'Debug Mode', level=Level.USER_DEV)
    
    class ImageDownloading(ConfigBase):
        convert_image = Setting[bool](True, 'Convert Image')
        preferable_format = Setting[str]('JXL', 'Converted Images Format')
        
        max_threads = RangeSetting[int](1, multiprocessing.cpu_count(), 1, multiprocessing.cpu_count(), 'Max Download Threads')
        image_update_every = RangeSetting[int](5, 100, 5, 10, 'Image Update Percentage', '%')
        
        
        PIL_SUPPORTED_EXT = Setting[dict[str, str]]({
            'JPG':  'JPEG',
            'JPEG': 'JPEG',
            'PNG':  'PNG',
            'GIF':  'GIF',
            'BMP':  'BMP',
            'WEBP': 'WEBP',
            'ICO':  'ICO',
            'TIFF': 'TIFF',
            'TIF':  'TIFF'
        }, 'Formats that PIL supports', level=Level.READ_ONLY)
        
    class UI(ConfigBase):
        class MangaViewer(ConfigBase):
            cull_height_multiplier = RangeSetting[float](1, 10, .1, 2, 'Cull Viewport Height Multiplier', level=Level.USER | Level.ADVANCED, setting_type=SettingType.PERFORMANCE | SettingType.COSMETIC, strongly_typed=False)
            cull_scene_cooldown = Setting[int](250, 'Scene Culling Minimum Cooldown', 'ms', level=Level.USER | Level.ADVANCED, setting_type=SettingType.PERFORMANCE)
        
    class Scrolling(ConfigBase):
        step = RangeSetting[int](10, 500, 20, 150, 'Step', 'px', setting_type=SettingType.COSMETIC | SettingType.QOL)
        step_duration = Setting[int](200, 'Step Duration', 'ms', setting_type=SettingType.COSMETIC | SettingType.QOL)
        alt_multiplier = Setting[int](8, 'Alt Step Multiplier', setting_type=SettingType.QOL)
        
        scale_multiplier = Setting[float](1.5, 'Step Scale Multiplier', level=Level.DEVELOPER, setting_type=SettingType.PERFORMANCE)
        
    class Dirs(StdDirConfigBase):
        STD_DIR = StdDirConfigBase.STD_DIR
        
        IMG = 'images'
        LOL = 'lol/kek/s.json'
        
        class CACHE(DirConfigBase):
            JSON = 'cache.json'
            
    class Enums(EnumsConfig):
        class Foo(En):
            BAR = a()
            BAR2 = a()

Config = Config_()
# Config.save('./setting.json')
# Config = Config.load('./setting.json')


from __future__ import annotations
from typing import Type, Protocol, Generic, TypeVar, get_args
from abc import ABC, abstractmethod
from enum import Flag, auto
import typing
from .type_enforcer import TypeEnforcer

# JSON serializable primitive types
type JsonPrimitiveTypes = str | int | float | bool | dict | list | None

# This is a type alias for all types that can be serialized to JSON
type JsonSerializableTypes = (
    JsonPrimitiveTypes | dict[typing.Any, typing.Any] | list[typing.Any]
)


class OutOfBoundsError(Exception): ...


class SettingValue(ABC):
    @abstractmethod
    def to_dict(self) -> dict: ...
    @abstractmethod
    def from_dict(self, dict_: dict): ...


class SettingValueProtocol(Protocol):
    @abstractmethod
    def to_dict(self) -> dict: ...
    @abstractmethod
    def from_dict(self, dict_: dict): ...


class Level(Flag):
    USER = auto()
    USER_DEV = auto()  # For users with dev_mode
    DEVELOPER = auto()  # For plugin developers
    ADVANCED = auto()
    READ_ONLY = auto()  # Cannot be modified
    READ_ONLY_USER = auto()  # Cannot be modifies by users, can be by code
    HIDDEN = auto()


class SettingType(Flag):
    PERFORMANCE = auto()
    COSMETIC = auto()
    QOL = auto()
    OTHER = auto()


T = TypeVar(
    "T", SettingValue, SettingValueProtocol, JsonSerializableTypes
)  # For nuitka:2.7 (2.8dev)


class SettingBase(Generic[T], TypeEnforcer[T]):
    JsonSerializableTypes = JsonSerializableTypes
    SettingValue = SettingValue
    SettingValueProtocol = SettingValueProtocol

    Level = Level
    SettingType = SettingType

    def __init__(
        self,
        *values: T,
        name: str = "",
        unit: str = "",  # TODO: Better units
        level: Level = Level.USER,
        setting_type: SettingType = SettingType.OTHER,
        description: str = "",
        strongly_typed: bool = True,
        try_parse_after_failure: bool = False,
    ):
        super().__init__(
            *values,
            name=name,
            strongly_typed=strongly_typed,
            try_parse_after_failure=try_parse_after_failure,
        )
        self.name = name
        self.unit = unit
        
        self.level = level
        self.setting_type = setting_type
        self.description = description
        
        self.is_active = True
        self.is_hide_inactive = False
        self.is_hidden = False

        self.dependent_settings = []
        self.dependent_settings_values: dict[T, Setting] = {}
        self.dependent_settings_func: list[tuple[typing.Callable, Setting]] = []

        self._boundaries_handler: typing.Callable = None
        
        self.__post_init__()

    def __post_init__(self):
        pass

    def add_dependent_setting(
        self,
        setting: Setting,
        value: T | None = None,
        value_fn: typing.Callable | None = None,
        hide_inactive: bool = False,
    ) -> typing.Self:
        """Add dependent setting to this setting.

        If value is provided, the setting will be activated when this setting's value is equal to the provided value.
        If value_func is provided, the setting will be activated when the function returns True for this setting's value.

        If hide_inactive is True, the setting will be hidden if it is not active.
        """
        if value:
            self.dependent_settings_values[value] = setting
        elif value_fn:
            self.dependent_settings_func.append((value_fn, setting))
        else:
            raise ValueError("Value or value_func should be provided")
        self.dependent_settings.append(setting)

        setting.set_hide_inactive(hide_inactive)
        if value in self.values or any([value_fn(value) for value in self.values]):
            setting.set_active(True)
        else:
            setting.set_active(False)
        return self
    
    def set_active(self, value: bool = True):
        self.is_active = value
        self.set_hidden(not value)
        return self

    def set_hide_inactive(self, value: bool = True):
        self.is_hide_inactive = value
        self.set_hidden(not self.is_active, respect_is_hide_inactive=value)

    def set_hidden(self, value: bool = True, respect_is_hide_inactive=True):
        """Set setting as hidden.

        If respect_is_hide_inactive is True, then if self.is_hide_inactive is True, the setting will be hidden if it is not active.
        Otherwise, the setting's hidden state will be set to the given value.
        """
        if self.is_hide_inactive and respect_is_hide_inactive:
            self.is_hidden = not self.is_active
        else:
            self.is_hidden = value
        return self

    def set_boundaries_handler(self, fn: typing.Callable) -> typing.Self:
        self._boundaries_handler = fn
        return self

    def set_boundaries_handler_for_value(
        self, fn: typing.Callable, value: T, overwrite_default=False
    ) -> typing.Self:
        return self

    def __set_name__(self, owner: Type, name: str):
        super().__set_name__(owner, name)
        self.name = (
            f"{owner.__name__}.{name} ({self.name})"
            if self.name
            else f"{owner.__name__}.{name}"
        )

    def __call__(self) -> T:
        return self.values
    
    def set_values(self, *values: T) -> typing.Self:
        if self.level is Level.READ_ONLY:
            raise PermissionError(f"Cannot modify a READ_ONLY setting: {self}")
        elif self._boundaries_handler and not all([self._boundaries_handler(value) for value in values]):
            raise OutOfBoundsError(
                f"Value {values} is out of bounds of given boundaries handler: {self}"
            )
        else:
            self.values = values
            self.enforce()

        return self
    
    def __set__(self, instance, *values: T) -> None:
        self.set_values(*values)
    
    def __str__(self) -> str:
        return f"{self.name}: {self._type}({str(self.values)})"  # type: ignore


class Setting(Generic[T], SettingBase[T]):
    def __init__(
        self,
        value: T,
        name: str = "",
        unit: str = "",  # TODO: Better units
        options=None,
        level: Level = Level.USER,
        setting_type: SettingType = SettingType.OTHER,
        description: str = "",
        strongly_typed: bool = True,
        try_parse_after_failure: bool = False,
    ):
        """Class to create setting in the Config. Is strongly typed.
        Create with `Setting[value_type](value)`.

        Value type must either be `Setting.SettingValue`, of `Setting.JsonSerializableTypes` type or use `Setting.SettingValueProtocol`

        By default Setting will not try parsing `value` into instance of `value_type`. Set `strongly_typed` to `False` to change it.

        Value of the class can be accessed by calling the instance:

        ```number = Setting[int](15, 'Some Number')
        number()   # returns a number.value (15)
        ```
        """
        super().__init__(
            value,
            name=name,
            unit=unit,
            level=level,
            setting_type=setting_type,
            description=description,
            strongly_typed=strongly_typed,
            try_parse_after_failure=try_parse_after_failure,
        )
        self.value = self.values[0]
        
        self.options = options

        self.default_value = self.value
    
    def __call__(self):
        return self.value
    
    def set_value(self, value: T) -> typing.Self:
        self.set_values(value)
        self.value = self.values[0]
        return self
    
    def __set__(self, instance, value: T) -> None:
        self.set_value(value)
    
    def __str__(self) -> str:
        return f"{self.name}: {self._type}({str(self.value)}{f' {self.unit}' if self.unit else ''})"  # type: ignore

    def to_dict(self) -> dict[str, JsonSerializableTypes]:
        """Convert setting to dictionary."""
        result = {
            "name": self.name,
            "options": self.options,
            "description": self.description,
            "type": self.setting_type,
        }
        if self._check_type(self._type, JsonSerializableTypes):
            result["value"] = self.value
        else:
            result["value"] = self.value.to_dict()

        return result


class VersionSetting(SettingBase[str | int]):
    def __post_init__(self):
        self.name = "Version"
        if len(self.values) == 1 and isinstance(self.values[1], str):
            self.version = self.values[0]
            self.major, self.minor, self.patch_status = self.version.split(".")
            self.major, self.minor = int(self.major), int(self.minor)
            self.patch, self.status = (
                -1,
                "" if "-" not in self.patch_status else self.patch_status.split("-"),
            )
            
            if not self.patch == -1:
                patch = ""
                for i, char in enumerate(self.patch_status):
                    if char.isnumeric():
                        patch += char
                    else:
                        self.patch = int(patch)
                        self.status = self.patch_status[i:]
                        break
        
        else:
            self.major, self.minor, *self.patch_status = self.values
            if isinstance(self.patch_status, list):
                self.patch, self.status = self.patch_status
            else:
                self.patch = int(self.patch_status)
                self.status = ''
            
            self.version = f'{self.major}.{self.minor}.{self.patch}{'-' + self.status if self.status else ''}'
            print(self.version)
    
    def __call__(self):
        return self.version
    
    def __str__(self) -> str:
        return f"{self.name}: {str(self.version)}"  # type: ignore


class RangeSetting(Generic[T], SettingBase[T]):
    def __init__(
        self,
        min_: T,
        max_: T,
        step: T,
        value: T,
        name: str = "",
        unit: str = "",  # TODO: Better units
        level: Level = Level.USER,
        setting_type: SettingType = SettingType.OTHER,
        description: str = "",
        strongly_typed: bool = True,
        try_parse_after_failure: bool = False,
    ):
        super().__init__(
            value, min_, max_, step,
            name=name,
            unit=unit,
            level=level,
            setting_type=setting_type,
            description=description,
            strongly_typed=strongly_typed,
            try_parse_after_failure=try_parse_after_failure,
        )
        self.value = value
        self.min = min_
        self.max = max_
        self.step_ = step
    
    def set_value(self, value: T) -> typing.Self:
        self.set_values(value)
        self.value = self.values[0]
        return self
    
    def __call__(self):
        return self.value
    
    def __set__(self, instance, value: T) -> None:
        self.set_value(value)
    
    def __str__(self) -> str:
        return f"{self.name}: {self._type}({str(self.value)}{f' {self.unit}' if self.unit else ''})"

    def step(self, steps: int = 1):
        value = self.value + self.step_ * steps
        if value < self.min or value > self.max:
            raise OutOfBoundsError(
                f"Cannot change value by {steps} steps. Calculated value: {value}, min: {self.min}, max: {self.max}"
            )
        self.set_value(value)


class CalculatedSetting(Setting[typing.Callable]):
    def __post_init__(self):
        self.fn = self.value
        return super().__post_init__()
    
    def __call__(self):
        return self.fn()


from __future__ import annotations
from pathlib import Path
import typing
import json
from .setting import Setting


class ConfigMeta(type):
    """Metaclass for Config to enable proper inheritance and access patterns."""

    def __new__(mcs, name: str, bases: tuple, namespace: dict) -> typing.Type:
        cls = super().__new__(mcs, name, bases, namespace)
        cls._config_name = name
        cls._at_class_creation(cls, name)
        
        # Process nested Config classes to make them proper attributes
        for key, value in cls.__dict__.items():
            cls._handle_key_value(cls, key, value)
            
        return cls
    
    @classmethod
    def _at_class_creation(cls, target_cls, name: str):
        target_cls._nested_configs = {}
        target_cls._settings = {}
    
    @classmethod
    def _handle_key_value(cls, target_cls, key, value):
        if isinstance(value, type) and isinstance(value, ConfigMeta):
            target_cls._nested_configs[key] = value()
        elif isinstance(value, Setting):
            target_cls._settings[key] = value


class ConfigBase(metaclass=ConfigMeta):
    """Base configuration class."""
    
    def __init__(self):
        self._nested_configs: dict[str, ConfigBase]
        self._settings: dict[str, Setting]
        
        for name, config_class in self._nested_configs.items():
            setattr(self, name, config_class)
    
    def get_all_configs(self) -> dict[str, ConfigBase]:
        return self._nested_configs
    
    def get_all_settings(self) -> dict[str, Setting]:
        return self._settings
        
    def to_dict(self) -> dict[str, typing.Any]:
        """Convert config to dictionary recursively."""
        result = {}

        # Process all attributes including inherited ones
        for key, value in vars(self.__class__).items():
            if key.startswith("_"):
                continue

            if isinstance(value, Setting):
                if hasattr(value(), "to_dict") and callable(value().to_dict):
                    result[key] = value().to_dict()
                else:
                    result[key] = value()
            elif issubclass(value, ConfigBase):
                result[key] = value().to_dict()

        return result

    # def to_dict_with_metadata(self) -> dict[str, ]:   # TODO: May be

    @classmethod
    def from_dict(cls, data: dict[str, typing.Any]) -> ConfigBase:
        """Create config instance from dictionary."""
        instance = cls()

        for key, value in data.items():
            attr = getattr(instance.__class__, key, None)

            if isinstance(attr, Setting):
                setattr(instance.__class__, key, value)
            elif isinstance(attr, ConfigBase) and isinstance(value, dict):
                # Recursively set nested config
                attr.__class__.from_dict(value)

        return instance

    def save(self, file_path: str | Path) -> None:
        """Save config to JSON file."""
        with open(file_path, "w") as f:
            json.dump(self.to_dict(), f, indent=4)

    @classmethod
    def load(cls, file_path: str | Path) -> ConfigBase:
        """Load config from JSON file."""
        with open(file_path, "r") as f:
            data = json.load(f)
        return cls.from_dict(data)

    def to_str(self, tabs) -> str:
        s = f"{self._config_name}:\n"
        for name, setting in self._settings.items():
            s += f'{'  '*tabs}{name}: {setting}\n'
        if self._settings:
            s += '\n'
        for config in self._nested_configs.values():
            s += '  '*tabs + config.to_str(tabs+1)
        return s
            
    def __str__(self) -> str:
        return self.to_str(1)


from enum import Enum, Flag, auto
from .config import ConfigMeta


En = Enum
Fl = Flag
a = au = auto

class EnumsConfigMeta(ConfigMeta):
    def _at_class_creation(cls, target_cls, name): ...
    def _handle_key_value(cls, target_cls, key, value): ...
    
class EnumsConfig(metaclass=EnumsConfigMeta): ...


from typing import (
    Union,
    _UnionGenericAlias,
    TypeVar,
    Any,
    Generic,
    Sequence,
    Set,
    MutableSet,
    MutableSequence,
    Mapping,
    MutableMapping,
    get_origin,
    get_args,
)
from types import UnionType, GenericAlias


class GenericMissingError(Exception): ...


class TypeEnforcerMeta(type):
    def __new__(cls, name, bases, namespace):
        cls = super().__new__(cls, name, bases, namespace)
        cls._class_name = name
        return cls

T = TypeVar('T')    # For nuitka:2.7 (2.8dev)

class TypeEnforcer(Generic[T], metaclass=TypeEnforcerMeta):
    _first__class_getitem__was_called = False
    _enforcer_cur_type = None  # Last type that was used in Enforcer
    _cur_type = None  # Last type that was used in Enforcer
    _enforcer_generic_type = None  # Type that was used in Enforcer in subclass
    _generic_type = None  # Type that was used in Enforcer in subclass

    def __class_getitem__(cls, item):
        """
        Called when Enforcer is used as a generic class.
        It checks if the class is a subclass of Enforcer and if it has a generic type.
        If it does, it replaces the type variable with the given type.
        """
        if TypeEnforcer._does_have_generic(item):  # Set the current type
            TypeEnforcer._enforcer_generic_type = item
        else:
            TypeEnforcer._enforcer_cur_type = item

        if (
            cls._generic_type
        ):  # If class has _generic_type, means it is getting its type (Child[int])
            new_type = TypeEnforcer._replace_typevar_with_type(cls._generic_type, {T.__name__: item})   # TODO: proper mapping
            if TypeEnforcer._does_have_generic(new_type):  # Save for next class
                TypeEnforcer._enforcer_generic_type = new_type
            else:
                TypeEnforcer._enforcer_cur_type = new_type
                cls._cur_type = new_type

        return super().__class_getitem__(item)

    @staticmethod
    def _replace_typevar_with_type(gen_type, type_map: dict[TypeVar, type]):
        """
        Replaces the type variable with the given type in the given generic type.
        """
        origin = get_origin(gen_type)
        if not origin:
            if isinstance(gen_type, TypeVar):
                return type_map.get(gen_type.__name__, gen_type)
            return gen_type
        
        args = list(get_args(gen_type))
        for i, t in enumerate(get_args(gen_type)):
            if isinstance(t, TypeVar):
                args[i] = type_map.get(t.__name__, gen_type)
            elif isinstance(t, _UnionGenericAlias):  # If it's generic and union
                args[i] = TypeEnforcer._replace_typevar_with_type(t, type_map)
            elif isinstance(t, GenericAlias):  # If its generic
                args[i] = TypeEnforcer._replace_typevar_with_type(t, type_map)
            elif (
                isinstance(t, UnionType)
            ):  # Strange situation if it's generic union, but its type is just UnionType, str | dict[str | int, T]
                args[i] = TypeEnforcer._replace_typevar_with_type(t, type_map)

        if origin is UnionType:
            return Union[*args]
        else:
            return origin[*args]

    @staticmethod
    def _does_have_generic(type_):
        if isinstance(type_, TypeVar):
            return True
        args = list(get_args(type_))
        for t in args:
            if isinstance(type_, TypeVar) or TypeEnforcer._does_have_generic(t):
                return True
        return False

    def __init_subclass__(cls) -> None:
        """
        Called when a subclass is created.
        It sets the current type and the generic type for the subclass.
        """
        cls._generic_type = TypeEnforcer._enforcer_generic_type
        cls._cur_type = TypeEnforcer._enforcer_cur_type
        TypeEnforcer._enforcer_generic_type = None
        TypeEnforcer._enforcer_cur_type = None
        return super().__init_subclass__()

    def __init__(
        self,
        *values: T,
        name: str = "",
        strongly_typed: bool = True,
        try_parse_after_failure=False,
    ):
        if not self._cur_type:
            raise TypeError(
                f"Class {self._class_name} should contain generic type: {self._class_name}[type]"
            )
        self._type = self._cur_type
        self.__class__._cur_type = None

        self.values = values
        self._name = name
        self.is_strongly_typed = strongly_typed
        self.is_try_parse_after_error = try_parse_after_failure
        
        self._attribute_owner = None
        self._attribute_name = ''
        
        self.enforce()
        
    def enforce(self):
        self.values = [self.enforce_type_on_value(value, self._type) for value in self.values]
        
    def enforce_type_on_value(self, value, type_=None, strongly=None):
        strongly = strongly or self.is_strongly_typed
        type_ = type_ or self._type
        if strongly:
            parsed = self._check_type(value, type_)
            val = value
        else:
            parsed, val = self._parse_check_type(
                value, type_, self.is_try_parse_after_error
            )
        
        if parsed:
            if val is not None:
                return val
        else:
            if strongly:
                raise TypeError(
                    f"Value of {self.full_name} must be of type {type_}\n"
                    f"Given {type(value)}({repr(value)})"
                )
            else:
                raise TypeError(
                    f"Value of {self.full_name} must be of type {type_} or corresponding values must be able to convert into corresponding types.\n"
                    f"Unable to convert {type(value)}({repr(value)}) in type {type_}"
                )

    def __set_name__(self, owner, name):
        self._attribute_owner = owner
        self._attribute_name = name
        
    @property
    def full_name(self):
        if not self._attribute_name:
            return self._name if self._name else self._class_name
        else:
            return f'{self._attribute_owner.__qualname__}.{self._attribute_name}'
        
    @classmethod
    def _check_type(cls, value, type_) -> bool:
        """
        Checks if the value is of the given type.

        Args:
            `value`: The value to check.
            `type_`: The type to check against.

        Returns:
            True if the value is of the given type, False otherwise.
        """
        orig = get_origin(type_)
        args = get_args(type_)

        # Is simple type (aka int, str)
        if not orig:
            return isinstance(value, type_)

        # Is union (aka int | str)
        elif orig is UnionType or orig is Union or type(orig) is UnionType:
            return any(cls._check_type(value, arg) for arg in args)

        # Is complex type (aka dict[str, int], list[dict[str, str | int]])
        else:
            if isinstance(value, orig):
                if isinstance(value, list) or issubclass(
                    orig, (Sequence, MutableSequence, Set, MutableSet)
                ):
                    return all(map(lambda x: cls._check_type(x, args[0]), value))

                if isinstance(value, dict) or issubclass(
                    orig, (Mapping, MutableMapping)
                ):
                    return all(
                        map(lambda x: cls._check_type(x, args[0]), value.keys())
                    ) and all(
                        map(lambda x: cls._check_type(x, args[1]), value.values())
                    )

            else:
                return False

        raise TypeError(f"Type '{type_}' is not supported yet")

    @classmethod
    def _parse_check_type(
        cls, value, type_, ignore_error=False
    ) -> tuple[bool, T | Any]:
        """
        Attempts to parse a value to a specified type, supporting complex and union types.

        This method checks whether a given value can be parsed into the specified type.
        If the value is not of the specified type, it attempts to convert it. The method
        supports simple types, union types, sequences, set, and mappings. If the conversion
        is successful, it returns the parsed value; otherwise, it returns the failed value,
        see `ignore_error`.

        Args:
            `value`: The value to be checked and potentially parsed.
            `type_`: The target type to parse the value into.
            `ignore_error` (bool, optional): If set to True, the method will attempt to parse
                                        all elements in sequences and mappings even if
                                        parsing of one element fails. Defaults to False.

        Returns:
            A tuple where the first element is a boolean indicating if the parsing was
            successful, and the second element is the parsed value or the original value
            if parsing was unsuccessful.
        """

        try:
            if cls._check_type(value=value, type_=type_):
                return True, value
        except TypeError:
            pass
        
        orig = get_origin(type_)
        args = get_args(type_)

        # Is simple type (aka int, str)
        if not orig:
            if not isinstance(value, type_):
                try:
                    return True, type_(value)
                except (ValueError, TypeError):
                    return False, value
            return True, value

        # Is union (aka int | str)
        elif orig is UnionType or orig is Union or type(orig) is UnionType:
            for arg in args:
                parsed, v = cls._parse_check_type(value, arg, ignore_error)
                if parsed:
                    break
            return parsed, v

        # Is complex type (aka dict[str, int], list[dict[str, str | int]])
        else:
            if isinstance(value, orig):
                if is_list := isinstance(value, list) or issubclass(
                    orig, (Sequence, MutableSequence, Set, MutableSet)
                ):
                    result_list = []
                    parsed = True
                    for v in value:
                        values_parsed, val = cls._parse_check_type(
                            v, args[0], ignore_error
                        )
                        result_list.append(val)
                        if not values_parsed:
                            parsed = False
                            if not ignore_error:
                                break
                    return parsed, result_list if is_list else type(value)(result_list)

                if isinstance(value, dict) or issubclass(
                    orig, (Mapping, MutableMapping)
                ):
                    result_dict = {}
                    parsed = True
                    for k, v in value.items():
                        keys_parsed, key = cls._parse_check_type(
                            k, args[0], ignore_error
                        )
                        values_parsed, val = cls._parse_check_type(
                            v, args[1], ignore_error
                        )
                        result_dict[key] = val
                        if not keys_parsed or not values_parsed:
                            parsed = False
                            if not ignore_error:
                                break
                    return parsed, result_dict

            else:
                return False, value


from .config import ConfigMeta, ConfigBase
from .directories_config import DirConfigMeta, StdDirConfigBase, DirConfigBase
from .enums_config import EnumsConfigMeta, EnumsConfig
from .setting import OutOfBoundsError, SettingValue, SettingValueProtocol, Level, SettingType, SettingBase, Setting, VersionSetting, RangeSetting, CalculatedSetting
from .type_enforcer import GenericMissingError, TypeEnforcerMeta, TypeEnforcer

__all__ = [
    'ConfigMeta', 
	'ConfigBase', 
	'DirConfigMeta', 
	'StdDirConfigBase', 
	'DirConfigBase', 
	'EnumsConfigMeta', 
	'EnumsConfig', 
	'OutOfBoundsError', 
	'SettingValue', 
	'SettingValueProtocol', 
	'Level', 
	'SettingType', 
	'SettingBase', 
	'Setting', 
	'VersionSetting', 
	'RangeSetting', 
	'CalculatedSetting', 
	'GenericMissingError', 
	'TypeEnforcerMeta', 
	'TypeEnforcer',
]


from src.easy_config_hub import Setting
import pytest


def test_init():
    s = Setting[int](69)


def test_type_init_error():
    with pytest.raises(TypeError):
        s = Setting(69)


def test_strict_type():
        s = Setting[int](69, strongly_typed=True)


def test_strict_type_error():
    with pytest.raises(TypeError):
        s = Setting[str](69, strongly_typed=True)
        

def test_parse_type():
    s = Setting[str](69, strongly_typed=False)


def test_parse_type_error():
    with pytest.raises(TypeError):
        s = Setting[int]("69 sixty nine", strongly_typed=False)


def test_union_strict_type():
    s = Setting[int | float](69.69)


def test_union_strict_type_error():
    with pytest.raises(TypeError):
        s = Setting[int | float]("69 sixty nine")


def test_union_parse_type():
    s = Setting[int | float]("69", strongly_typed=False)
    assert s() == 69
    assert type(s()) is int


def test_union_parse_type_error():
    with pytest.raises(TypeError):
        s = Setting[int | float]("69 sixty nine", strongly_typed=False)


def test_union_parse_right_second_type():
    s = Setting[str | int](69, strongly_typed=False)
    assert type(s()) is int


def test_list_strict_type():
    s = Setting[list[str]](["sdkfj", "sdfjs"])


def test_list_strict_type_error():
    with pytest.raises(TypeError):
        s = Setting[list[str]](["sdkfj", "sdfjs", 69, "asdf"])


def test_list_parse_type():
    s = Setting[list[float]](["420", "69"], strongly_typed=False)
    assert s() == [420.0, 69.0]
    assert type(s()[0]) is float
    assert type(s()[1]) is float


def test_list_parse_type_error():
    with pytest.raises(TypeError):
        s = Setting[list[int]](["420", "69 sixty nine", "1"], strongly_typed=False)


def test_dict_strict_type():
    s = Setting[dict[str, int]]({"sdkfj": 69, "sdfjs": 420})


def test_dict_strict_type_error():
    with pytest.raises(TypeError):
        s = Setting[dict[str, int]]({"sdkfj": 420, 69: 96, "asdf": 23})

def test_dict_parse_type():
    s = Setting[dict[float, int]]({"420": "024", "69": "96"}, strongly_typed=False)
    assert s() == {420.0: 24, 69.0: 96}
    assert type(list(s().keys())[0]) is float
    assert type(list(s().values())[0]) is int
    assert type(list(s().keys())[1]) is float
    assert type(list(s().values())[1]) is int


def test_dict_parse_type_error():
    with pytest.raises(TypeError):
        s = Setting[dict[int, int]](
            {"420": "024", "69 sixty nine": "96", "1": 1}, strongly_typed=False
        )


def test_really_complex_union_strict_type():
    s = Setting[list[dict[str, list[int | str]]] | int | float](
        [{"lol": [69, "69"], "kek": [420]}, {"lol": [69, 69], "kek": ["420"]}]
    )


# def test_complex_parse_type_after_error():
#     try:
#         s = Setting[list[int]](['420', '69 sixty nine', '1'], strongly_typed=False, try_parse_after_failure=True)
#     except TypeError:
#         pass
#     finally:
#         assert s()[0] == 420
#         assert s()[1] == 2
#         assert s()[2] == 1
